// LogBlendImpl.cpp : Implementation of CLogoBlendImpl

#include "stdafx.h"
#include <atlstr.h>
#include "LogoBlendImpl.h"

// The following Macros deal with BMP padding.
#define VSWIDTHBYTES(bits) ((DWORD)(((bits)+31) & (~31)) / 8)
#define VSDIBWIDTHBYTES(bi) (DWORD)VSWIDTHBYTES((DWORD)(bi).biWidth * (DWORD)(bi).biBitCount)


///////////////////////////////////////////////////////////////////////////////
// IUVideoTransformer::Transform method.
//
// Purpose:    Called by the ULiveServer for every video frame streamed to the
//             users.
//
// Parameters: [in]  long               - ID of the live resources setup 
//                                        with Live Server Configurator.
//             [in]  U_VideoFrame*      - Video Frame.
//             [out] U_VideoFrameData** - Transformed Video Frame.
///////////////////////////////////////////////////////////////////////////////
STDMETHODIMP CLogoBlendImpl::Transform(long lResourceID, U_VideoFrame *pInFrame, 
									   U_VideoFrameData **pOutFrame)
{
	HRESULT hr = S_OK;

	try
	{
		// Allocate destination frame buffer.
		*pOutFrame = (U_VideoFrameData*)CoTaskMemAlloc(sizeof(U_VideoFrameData));
		(*pOutFrame)->pFrame = (BYTE*)CoTaskMemAlloc(pInFrame->Frame.nLength);

		// Size of transformed frame must be equal to the size of source frame.
		(*pOutFrame)->nLength = pInFrame->Frame.nLength;

		// For COM-compliance we need to copy frame.
		memcpy((*pOutFrame)->pFrame, pInFrame->Frame.pFrame, (*pOutFrame)->nLength);
		
		// Do blending. Subtype of received video frame is always RGB24, i.e., 
		// pInFrame->subtype is equal to MEDIASUBTYPE_RGB24.
		BITMAPINFOHEADER bmh;
		ZeroMemory(&bmh, sizeof(BITMAPINFOHEADER));
		bmh.biWidth    = pInFrame->lWidth;
		bmh.biHeight   = pInFrame->lHeight;
		bmh.biBitCount = 24;
		bmh.biSize     = sizeof(BITMAPINFOHEADER);
		bmh.biPlanes   = 1;

		TransparentBlend((*pOutFrame)->pFrame, bmh, m_pLogoBits, m_LogoBMPHeader, 3);
	}
	catch (...)
	{
		hr = E_UNEXPECTED;
	}

	return hr;
}

bool CLogoBlendImpl::LoadLogo(LPCTSTR szLogoFileName)
{
	CString sFileName;
	GetModuleFileName(_AtlBaseModule.m_hInst, sFileName.GetBuffer(MAX_PATH), MAX_PATH);
	sFileName.ReleaseBuffer();

	int nPos  = sFileName.ReverseFind(_T('\\'));
	sFileName = sFileName.Left(nPos + 1);
	sFileName += szLogoFileName;

	return LoadBitmapFile(sFileName, &m_LogoBMPHeader, &m_pLogoBits);
}

// This function overlays source bitmap on top of the destination bitmap, blending intensities of the
// pixels according to nLogoWeight parameter. The more nLogoWeight, the more transparent will source
// bitmap look. Source bitmap can only be 24-bit, destination bitmap can be 24-bit or 8-bit (greyscale)
bool CLogoBlendImpl::TransparentBlend(BYTE *pDestBytes, BITMAPINFOHEADER &Destbmh, 
					                  BYTE *pSourceBytes, BITMAPINFOHEADER &Sourcebmh, 
									  long nLogoWeight)
{
	if (nLogoWeight <= 0)
	{
		return false;
	}

	if ((Destbmh.biBitCount != 24)&&(Destbmh.biBitCount != 8))
	{
		return false;
	}

	if (Sourcebmh.biBitCount != 24)
	{
		return false;
	}

	if (Sourcebmh.biHeight >= Destbmh.biHeight || Sourcebmh.biWidth >= Destbmh.biWidth)
	{
		return false;
	}

	LONG nSourceRowWidth = VSDIBWIDTHBYTES(Sourcebmh);
	LONG nDestRowWidth   = VSDIBWIDTHBYTES(Destbmh);

	BYTE *pDestPointer   = pDestBytes + nDestRowWidth * (Destbmh.biHeight - Sourcebmh.biHeight);
	BYTE *pSourcePointer = pSourceBytes;
	
	for (LONG i = 0; i < Sourcebmh.biHeight; i++)
	{
		LONG nSourceRowWidthNoPadding = Sourcebmh.biWidth*3;
		for (LONG j = 0; j < nSourceRowWidthNoPadding; j++)
		{
			BYTE nSourceAdditionIntensity = (BYTE)((Destbmh.biBitCount == 24)? pSourcePointer[j]/nLogoWeight:
				 (pSourcePointer[j] + pSourcePointer[j + 1] + pSourcePointer[j + 2])/3/nLogoWeight);

			LONG k = (Destbmh.biBitCount == 24)? j : j/3;
			
			if (pDestPointer[k] + nSourceAdditionIntensity < 255)
			{
				pDestPointer[k] = (pDestPointer[k] + nSourceAdditionIntensity);
			}
			else
			{
				pDestPointer[k] = 255;
			}
			
			if (Destbmh.biBitCount == 8)
			{
				j += 2;
				continue;
			}
		}

		pDestPointer   += nDestRowWidth;
		pSourcePointer += nSourceRowWidth;
	}

	return true;
}

// Loads bitmap from file. BMP files only.
bool CLogoBlendImpl::LoadBitmapFile(LPCTSTR szFileName, BITMAPINFOHEADER *pBmh, PBYTE *ppBuffer)
{
	CAutoHandle hFile = CreateFile(szFileName, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE)
	{
		return false;
	}

	DWORD Read;
	BITMAPFILEHEADER bmfHeader;
	if (!ReadFile(hFile, &bmfHeader, sizeof(BITMAPFILEHEADER), &Read, NULL))
	{
		return false;
	}

	if (bmfHeader.bfType != ((WORD) ('M' << 8) | 'B'))
	{
		return false;
	}

	if (!ReadFile(hFile, pBmh, sizeof(BITMAPINFOHEADER), &Read, NULL))
	{
		return false;
	}

	if (!SetFilePointer(hFile, bmfHeader.bfOffBits, 0, FILE_BEGIN))
	{
		return false;
	}

	*ppBuffer = new BYTE[pBmh->biSizeImage];
	if (!ReadFile(hFile, *ppBuffer, pBmh->biSizeImage , &Read, NULL))
	{
		delete *ppBuffer;
		return false;
	}

	return true;
}
